import logging

from grpclib.events import RecvRequest
from grpclib.server import Server

from cilroy.controller import CilroyController
from cilroy.service import CilroyService


class CilroyServer:
    def __init__(
        self,
        controller: CilroyController,
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> None:
        self._service = CilroyService(controller)
        self._logger = logger

    async def _on_request(self, event: RecvRequest) -> None:
        self._logger.debug(f'Handling "{event.method_name}"...')

    async def run(self, *args, **kwargs) -> None:
        server = Server([self._service])
        await server.start(*args, **kwargs)
        self._logger.info("Server started.")
        try:
            await server.wait_closed()
        finally:
            self._logger.info("Server stopped.")



from datetime import datetime, time, timedelta, timezone
from math import ceil
from types import TracebackType
from typing import (
    AsyncIterable,
    AsyncIterator,
    Generic,
    Iterable,
    MutableMapping,
    Optional,
    Type,
    TypeVar,
    Union,
)
from uuid import uuid4

from aiostream.stream import chain, iterate, preserve, take

T = TypeVar("T")


class CachingAsyncIterable(AsyncIterable[T], Generic[T]):
    _iterable: AsyncIterable[T]
    _cache: MutableMapping[str, T]
    _prefix: str
    _watermark: int

    def __init__(
        self,
        iterable: AsyncIterable[T],
        cache: Optional[MutableMapping[str, T]] = None,
        prefix: Optional[str] = None,
    ):
        self._iterable = iterable
        self._cache = cache if cache is not None else {}
        self._prefix = prefix if prefix is not None else uuid4().hex
        self._watermark = 0

    def _make_key(self, i: int) -> str:
        return f"{self._prefix}-{i}"

    async def __aiter__(self) -> AsyncIterator[T]:
        for i in range(self._watermark):
            key = self._make_key(i)
            yield self._cache[key]
        async for item in self._iterable:
            key = self._make_key(self._watermark)
            self._cache[key] = item
            self._watermark += 1
            yield item

    def __enter__(self) -> "CachingAsyncIterable[T]":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        for i in range(self._watermark):
            key = self._make_key(i)
            del self._cache[key]
        self._watermark = 0
        return None


async def batches(
    iterable: Union[Iterable[T], AsyncIterable[T]],
    n: Optional[int],
) -> AsyncIterable[AsyncIterable[T]]:
    iterable = next(iterate(iterable)._generator)

    if n is None:
        yield iterable
        return

    async for first in iterable:

        async def afirst():
            yield first

        batch = chain(afirst(), take(preserve(iterable), n - 1))
        yield next(batch._generator)


def next_time(base: datetime, interval: timedelta) -> datetime:
    utcbase = base.astimezone(timezone.utc)
    utcnow = datetime.now(timezone.utc)
    return utcbase + interval * ceil((utcnow - utcbase) / interval)


def seconds_until(dt: datetime) -> float:
    return (dt - datetime.now(timezone.utc)).total_seconds()


def utcmidnight() -> datetime:
    now = datetime.now(timezone.utc)
    return datetime.combine(now.date(), time(), tzinfo=timezone.utc)



from abc import ABC, abstractmethod
from typing import AsyncIterable, Awaitable, Callable, Dict, Tuple
from uuid import UUID

from kilroy_server_py_utils import Categorizable, classproperty, normalize


class PostScheduler(Categorizable, ABC):
    @classproperty
    def category(cls) -> str:
        name: str = cls.__name__
        return normalize(name.removesuffix("PostScheduler"))

    @abstractmethod
    def wait(self) -> AsyncIterable[None]:
        pass



from abc import ABC, abstractmethod
from typing import (
    AsyncIterable,
    Awaitable,
    Callable,
    Collection,
    Dict,
    Set,
)
from uuid import UUID

from kilroy_server_py_utils import Categorizable, classproperty, normalize


class ScoreScheduler(Categorizable, ABC):
    @classproperty
    def category(cls) -> str:
        name: str = cls.__name__
        return normalize(name.removesuffix("ScoreScheduler"))

    @abstractmethod
    def wait(self) -> AsyncIterable[None]:
        pass



import json
from typing import AsyncIterator, Dict, Tuple

import aiostream
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase
from grpclib.server import Stream

from cilroy.controller import CilroyController
from cilroy.messages import *


class CilroyServiceBase(ServiceBase):
    async def get_face_metadata(
        self, get_face_metadata_request: "GetFaceMetadataRequest"
    ) -> "GetFaceMetadataResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_module_metadata(
        self, get_module_metadata_request: "GetModuleMetadataRequest"
    ) -> "GetModuleMetadataResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_face_post_schema(
        self, get_face_post_schema_request: "GetFacePostSchemaRequest"
    ) -> "GetFacePostSchemaResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_module_post_schema(
        self, get_module_post_schema_request: "GetModulePostSchemaRequest"
    ) -> "GetModulePostSchemaResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_controller_status(
        self, get_controller_status_request: "GetControllerStatusRequest"
    ) -> "GetControllerStatusResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def watch_controller_status(
        self, watch_controller_status_request: "WatchControllerStatusRequest"
    ) -> AsyncIterator["WatchControllerStatusResponse"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_face_status(
        self, get_face_status_request: "GetFaceStatusRequest"
    ) -> "GetFaceStatusResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def watch_face_status(
        self, watch_face_status_request: "WatchFaceStatusRequest"
    ) -> AsyncIterator["WatchFaceStatusResponse"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_module_status(
        self, get_module_status_request: "GetModuleStatusRequest"
    ) -> "GetModuleStatusResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def watch_module_status(
        self, watch_module_status_request: "WatchModuleStatusRequest"
    ) -> AsyncIterator["WatchModuleStatusResponse"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_training_status(
        self, get_training_status_request: "GetTrainingStatusRequest"
    ) -> "GetTrainingStatusResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def watch_training_status(
        self, watch_training_status_request: "WatchTrainingStatusRequest"
    ) -> AsyncIterator["WatchTrainingStatusResponse"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_controller_config_schema(
        self,
        get_controller_config_schema_request: "GetControllerConfigSchemaRequest",
    ) -> "GetControllerConfigSchemaResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_controller_config(
        self, get_controller_config_request: "GetControllerConfigRequest"
    ) -> "GetControllerConfigResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def watch_controller_config(
        self, watch_controller_config_request: "WatchControllerConfigRequest"
    ) -> AsyncIterator["WatchControllerConfigResponse"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def set_controller_config(
        self, set_controller_config_request: "SetControllerConfigRequest"
    ) -> "SetControllerConfigResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_face_config_schema(
        self, get_face_config_schema_request: "GetFaceConfigSchemaRequest"
    ) -> "GetFaceConfigSchemaResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_face_config(
        self, get_face_config_request: "GetFaceConfigRequest"
    ) -> "GetFaceConfigResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def watch_face_config(
        self, watch_face_config_request: "WatchFaceConfigRequest"
    ) -> AsyncIterator["WatchFaceConfigResponse"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def set_face_config(
        self, set_face_config_request: "SetFaceConfigRequest"
    ) -> "SetFaceConfigResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_module_config_schema(
        self, get_module_config_schema_request: "GetModuleConfigSchemaRequest"
    ) -> "GetModuleConfigSchemaResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_module_config(
        self, get_module_config_request: "GetModuleConfigRequest"
    ) -> "GetModuleConfigResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def watch_module_config(
        self, watch_module_config_request: "WatchModuleConfigRequest"
    ) -> AsyncIterator["WatchModuleConfigResponse"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def set_module_config(
        self, set_module_config_request: "SetModuleConfigRequest"
    ) -> "SetModuleConfigResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def train_offline(
        self, train_offline_request: "TrainOfflineRequest"
    ) -> "TrainOfflineResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def train_online(
        self, train_online_request: "TrainOnlineRequest"
    ) -> "TrainOnlineResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def stop_training(
        self, stop_training_request: "StopTrainingRequest"
    ) -> "StopTrainingResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_module_metrics_config(
        self,
        get_module_metrics_config_request: "GetModuleMetricsConfigRequest",
    ) -> "GetModuleMetricsConfigResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_module_metrics(
        self, get_module_metrics_request: "GetModuleMetricsRequest"
    ) -> "GetModuleMetricsResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def watch_module_metrics(
        self, watch_module_metrics_request: "WatchModuleMetricsRequest"
    ) -> AsyncIterator["WatchModuleMetricsResponse"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def watch_all(
        self, watch_all_request: "WatchAllRequest"
    ) -> AsyncIterator["WatchAllResponse"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def reset_controller(
        self, reset_controller_request: "ResetControllerRequest"
    ) -> "ResetControllerResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def reset_face(
        self, reset_face_request: "ResetFaceRequest"
    ) -> "ResetFaceResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def reset_module(
        self, reset_module_request: "ResetModuleRequest"
    ) -> "ResetModuleResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_feed(
        self, get_feed_request: "GetFeedRequest"
    ) -> "GetFeedResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def watch_feed(
        self, watch_feed_request: "WatchFeedRequest"
    ) -> AsyncIterator["WatchFeedResponse"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_get_face_metadata(
        self,
        stream: "grpclib.server.Stream[GetFaceMetadataRequest, GetFaceMetadataResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_face_metadata(request)
        await stream.send_message(response)

    async def __rpc_get_module_metadata(
        self,
        stream: "grpclib.server.Stream[GetModuleMetadataRequest, GetModuleMetadataResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_module_metadata(request)
        await stream.send_message(response)

    async def __rpc_get_face_post_schema(
        self,
        stream: "grpclib.server.Stream[GetFacePostSchemaRequest, GetFacePostSchemaResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_face_post_schema(request)
        await stream.send_message(response)

    async def __rpc_get_module_post_schema(
        self,
        stream: "grpclib.server.Stream[GetModulePostSchemaRequest, GetModulePostSchemaResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_module_post_schema(request)
        await stream.send_message(response)

    async def __rpc_get_controller_status(
        self,
        stream: "grpclib.server.Stream[GetControllerStatusRequest, GetControllerStatusResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_controller_status(request)
        await stream.send_message(response)

    async def __rpc_watch_controller_status(
        self,
        stream: "grpclib.server.Stream[WatchControllerStatusRequest, WatchControllerStatusResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.watch_controller_status,
            stream,
            request,
        )

    async def __rpc_get_face_status(
        self,
        stream: "grpclib.server.Stream[GetFaceStatusRequest, GetFaceStatusResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_face_status(request)
        await stream.send_message(response)

    async def __rpc_watch_face_status(
        self,
        stream: "grpclib.server.Stream[WatchFaceStatusRequest, WatchFaceStatusResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.watch_face_status,
            stream,
            request,
        )

    async def __rpc_get_module_status(
        self,
        stream: "grpclib.server.Stream[GetModuleStatusRequest, GetModuleStatusResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_module_status(request)
        await stream.send_message(response)

    async def __rpc_watch_module_status(
        self,
        stream: "grpclib.server.Stream[WatchModuleStatusRequest, WatchModuleStatusResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.watch_module_status,
            stream,
            request,
        )

    async def __rpc_get_training_status(
        self,
        stream: "grpclib.server.Stream[GetTrainingStatusRequest, GetTrainingStatusResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_training_status(request)
        await stream.send_message(response)

    async def __rpc_watch_training_status(
        self,
        stream: "grpclib.server.Stream[WatchTrainingStatusRequest, WatchTrainingStatusResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.watch_training_status,
            stream,
            request,
        )

    async def __rpc_get_controller_config_schema(
        self,
        stream: "grpclib.server.Stream[GetControllerConfigSchemaRequest, GetControllerConfigSchemaResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_controller_config_schema(request)
        await stream.send_message(response)

    async def __rpc_get_controller_config(
        self,
        stream: "grpclib.server.Stream[GetControllerConfigRequest, GetControllerConfigResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_controller_config(request)
        await stream.send_message(response)

    async def __rpc_watch_controller_config(
        self,
        stream: "grpclib.server.Stream[WatchControllerConfigRequest, WatchControllerConfigResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.watch_controller_config,
            stream,
            request,
        )

    async def __rpc_set_controller_config(
        self,
        stream: "grpclib.server.Stream[SetControllerConfigRequest, SetControllerConfigResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.set_controller_config(request)
        await stream.send_message(response)

    async def __rpc_get_face_config_schema(
        self,
        stream: "grpclib.server.Stream[GetFaceConfigSchemaRequest, GetFaceConfigSchemaResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_face_config_schema(request)
        await stream.send_message(response)

    async def __rpc_get_face_config(
        self,
        stream: "grpclib.server.Stream[GetFaceConfigRequest, GetFaceConfigResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_face_config(request)
        await stream.send_message(response)

    async def __rpc_watch_face_config(
        self,
        stream: "grpclib.server.Stream[WatchFaceConfigRequest, WatchFaceConfigResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.watch_face_config,
            stream,
            request,
        )

    async def __rpc_set_face_config(
        self,
        stream: "grpclib.server.Stream[SetFaceConfigRequest, SetFaceConfigResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.set_face_config(request)
        await stream.send_message(response)

    async def __rpc_get_module_config_schema(
        self,
        stream: "grpclib.server.Stream[GetModuleConfigSchemaRequest, GetModuleConfigSchemaResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_module_config_schema(request)
        await stream.send_message(response)

    async def __rpc_get_module_config(
        self,
        stream: "grpclib.server.Stream[GetModuleConfigRequest, GetModuleConfigResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_module_config(request)
        await stream.send_message(response)

    async def __rpc_watch_module_config(
        self,
        stream: "grpclib.server.Stream[WatchModuleConfigRequest, WatchModuleConfigResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.watch_module_config,
            stream,
            request,
        )

    async def __rpc_set_module_config(
        self,
        stream: "grpclib.server.Stream[SetModuleConfigRequest, SetModuleConfigResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.set_module_config(request)
        await stream.send_message(response)

    async def __rpc_train_offline(
        self,
        stream: "grpclib.server.Stream[TrainOfflineRequest, TrainOfflineResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.train_offline(request)
        await stream.send_message(response)

    async def __rpc_train_online(
        self,
        stream: "grpclib.server.Stream[TrainOnlineRequest, TrainOnlineResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.train_online(request)
        await stream.send_message(response)

    async def __rpc_stop_training(
        self,
        stream: "grpclib.server.Stream[StopTrainingRequest, StopTrainingResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.stop_training(request)
        await stream.send_message(response)

    async def __rpc_get_module_metrics_config(
        self,
        stream: "grpclib.server.Stream[GetModuleMetricsConfigRequest, GetModuleMetricsConfigResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_module_metrics_config(request)
        await stream.send_message(response)

    async def __rpc_get_module_metrics(
        self,
        stream: "grpclib.server.Stream[GetModuleMetricsRequest, GetModuleMetricsResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_module_metrics(request)
        await stream.send_message(response)

    async def __rpc_watch_module_metrics(
        self,
        stream: "grpclib.server.Stream[WatchModuleMetricsRequest, WatchModuleMetricsResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.watch_module_metrics,
            stream,
            request,
        )

    async def __rpc_watch_all(
        self,
        stream: "grpclib.server.Stream[WatchAllRequest, WatchAllResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.watch_all,
            stream,
            request,
        )

    async def __rpc_reset_controller(
        self,
        stream: "grpclib.server.Stream[ResetControllerRequest, ResetControllerResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.reset_controller(request)
        await stream.send_message(response)

    async def __rpc_reset_face(
        self,
        stream: "grpclib.server.Stream[ResetFaceRequest, ResetFaceResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.reset_face(request)
        await stream.send_message(response)

    async def __rpc_reset_module(
        self,
        stream: "grpclib.server.Stream[ResetModuleRequest, ResetModuleResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.reset_module(request)
        await stream.send_message(response)

    async def __rpc_get_feed(
        self, stream: "grpclib.server.Stream[GetFeedRequest, GetFeedResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_feed(request)
        await stream.send_message(response)

    async def __rpc_watch_feed(
        self,
        stream: "grpclib.server.Stream[WatchFeedRequest, WatchFeedResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.watch_feed,
            stream,
            request,
        )

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/kilroy.cilroy.v1alpha.CilroyService/GetFaceMetadata": grpclib.const.Handler(
                self.__rpc_get_face_metadata,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetFaceMetadataRequest,
                GetFaceMetadataResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetModuleMetadata": grpclib.const.Handler(
                self.__rpc_get_module_metadata,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetModuleMetadataRequest,
                GetModuleMetadataResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetFacePostSchema": grpclib.const.Handler(
                self.__rpc_get_face_post_schema,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetFacePostSchemaRequest,
                GetFacePostSchemaResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetModulePostSchema": grpclib.const.Handler(
                self.__rpc_get_module_post_schema,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetModulePostSchemaRequest,
                GetModulePostSchemaResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetControllerStatus": grpclib.const.Handler(
                self.__rpc_get_controller_status,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetControllerStatusRequest,
                GetControllerStatusResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/WatchControllerStatus": grpclib.const.Handler(
                self.__rpc_watch_controller_status,
                grpclib.const.Cardinality.UNARY_STREAM,
                WatchControllerStatusRequest,
                WatchControllerStatusResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetFaceStatus": grpclib.const.Handler(
                self.__rpc_get_face_status,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetFaceStatusRequest,
                GetFaceStatusResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/WatchFaceStatus": grpclib.const.Handler(
                self.__rpc_watch_face_status,
                grpclib.const.Cardinality.UNARY_STREAM,
                WatchFaceStatusRequest,
                WatchFaceStatusResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetModuleStatus": grpclib.const.Handler(
                self.__rpc_get_module_status,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetModuleStatusRequest,
                GetModuleStatusResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/WatchModuleStatus": grpclib.const.Handler(
                self.__rpc_watch_module_status,
                grpclib.const.Cardinality.UNARY_STREAM,
                WatchModuleStatusRequest,
                WatchModuleStatusResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetTrainingStatus": grpclib.const.Handler(
                self.__rpc_get_training_status,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetTrainingStatusRequest,
                GetTrainingStatusResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/WatchTrainingStatus": grpclib.const.Handler(
                self.__rpc_watch_training_status,
                grpclib.const.Cardinality.UNARY_STREAM,
                WatchTrainingStatusRequest,
                WatchTrainingStatusResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetControllerConfigSchema": grpclib.const.Handler(
                self.__rpc_get_controller_config_schema,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetControllerConfigSchemaRequest,
                GetControllerConfigSchemaResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetControllerConfig": grpclib.const.Handler(
                self.__rpc_get_controller_config,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetControllerConfigRequest,
                GetControllerConfigResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/WatchControllerConfig": grpclib.const.Handler(
                self.__rpc_watch_controller_config,
                grpclib.const.Cardinality.UNARY_STREAM,
                WatchControllerConfigRequest,
                WatchControllerConfigResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/SetControllerConfig": grpclib.const.Handler(
                self.__rpc_set_controller_config,
                grpclib.const.Cardinality.UNARY_UNARY,
                SetControllerConfigRequest,
                SetControllerConfigResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetFaceConfigSchema": grpclib.const.Handler(
                self.__rpc_get_face_config_schema,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetFaceConfigSchemaRequest,
                GetFaceConfigSchemaResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetFaceConfig": grpclib.const.Handler(
                self.__rpc_get_face_config,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetFaceConfigRequest,
                GetFaceConfigResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/WatchFaceConfig": grpclib.const.Handler(
                self.__rpc_watch_face_config,
                grpclib.const.Cardinality.UNARY_STREAM,
                WatchFaceConfigRequest,
                WatchFaceConfigResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/SetFaceConfig": grpclib.const.Handler(
                self.__rpc_set_face_config,
                grpclib.const.Cardinality.UNARY_UNARY,
                SetFaceConfigRequest,
                SetFaceConfigResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetModuleConfigSchema": grpclib.const.Handler(
                self.__rpc_get_module_config_schema,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetModuleConfigSchemaRequest,
                GetModuleConfigSchemaResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetModuleConfig": grpclib.const.Handler(
                self.__rpc_get_module_config,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetModuleConfigRequest,
                GetModuleConfigResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/WatchModuleConfig": grpclib.const.Handler(
                self.__rpc_watch_module_config,
                grpclib.const.Cardinality.UNARY_STREAM,
                WatchModuleConfigRequest,
                WatchModuleConfigResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/SetModuleConfig": grpclib.const.Handler(
                self.__rpc_set_module_config,
                grpclib.const.Cardinality.UNARY_UNARY,
                SetModuleConfigRequest,
                SetModuleConfigResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/TrainOffline": grpclib.const.Handler(
                self.__rpc_train_offline,
                grpclib.const.Cardinality.UNARY_UNARY,
                TrainOfflineRequest,
                TrainOfflineResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/TrainOnline": grpclib.const.Handler(
                self.__rpc_train_online,
                grpclib.const.Cardinality.UNARY_UNARY,
                TrainOnlineRequest,
                TrainOnlineResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/StopTraining": grpclib.const.Handler(
                self.__rpc_stop_training,
                grpclib.const.Cardinality.UNARY_UNARY,
                StopTrainingRequest,
                StopTrainingResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetModuleMetricsConfig": grpclib.const.Handler(
                self.__rpc_get_module_metrics_config,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetModuleMetricsConfigRequest,
                GetModuleMetricsConfigResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetModuleMetrics": grpclib.const.Handler(
                self.__rpc_get_module_metrics,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetModuleMetricsRequest,
                GetModuleMetricsResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/WatchModuleMetrics": grpclib.const.Handler(
                self.__rpc_watch_module_metrics,
                grpclib.const.Cardinality.UNARY_STREAM,
                WatchModuleMetricsRequest,
                WatchModuleMetricsResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/WatchAll": grpclib.const.Handler(
                self.__rpc_watch_all,
                grpclib.const.Cardinality.UNARY_STREAM,
                WatchAllRequest,
                WatchAllResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/ResetController": grpclib.const.Handler(
                self.__rpc_reset_controller,
                grpclib.const.Cardinality.UNARY_UNARY,
                ResetControllerRequest,
                ResetControllerResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/ResetFace": grpclib.const.Handler(
                self.__rpc_reset_face,
                grpclib.const.Cardinality.UNARY_UNARY,
                ResetFaceRequest,
                ResetFaceResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/ResetModule": grpclib.const.Handler(
                self.__rpc_reset_module,
                grpclib.const.Cardinality.UNARY_UNARY,
                ResetModuleRequest,
                ResetModuleResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/GetFeed": grpclib.const.Handler(
                self.__rpc_get_feed,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetFeedRequest,
                GetFeedResponse,
            ),
            "/kilroy.cilroy.v1alpha.CilroyService/WatchFeed": grpclib.const.Handler(
                self.__rpc_watch_feed,
                grpclib.const.Cardinality.UNARY_STREAM,
                WatchFeedRequest,
                WatchFeedResponse,
            ),
        }


class CilroyService(CilroyServiceBase):
    def __init__(self, controller: CilroyController) -> None:
        super().__init__()
        self._controller = controller

    async def get_face_metadata(
        self, get_face_metadata_request: "GetFaceMetadataRequest"
    ) -> "GetFaceMetadataResponse":
        metadata = await self._controller.get_face_metadata()
        return GetFaceMetadataResponse(
            key=metadata.key,
            description=metadata.description,
        )

    async def get_module_metadata(
        self, get_module_metadata_request: "GetModuleMetadataRequest"
    ) -> "GetModuleMetadataResponse":
        metadata = await self._controller.get_module_metadata()
        return GetModuleMetadataResponse(
            key=metadata.key,
            description=metadata.description,
        )

    async def get_face_post_schema(
        self, get_face_post_schema_request: "GetFacePostSchemaRequest"
    ) -> "GetFacePostSchemaResponse":
        schema = await self._controller.get_face_post_schema()
        return GetFacePostSchemaResponse().from_dict({"schema": schema.json()})

    async def get_module_post_schema(
        self, get_module_post_schema_request: "GetModulePostSchemaRequest"
    ) -> "GetModulePostSchemaResponse":
        schema = await self._controller.get_module_post_schema()
        return GetModulePostSchemaResponse().from_dict(
            {"schema": schema.json()}
        )

    async def get_controller_status(
        self, get_controller_status_request: "GetControllerStatusRequest"
    ) -> "GetControllerStatusResponse":
        ready = await self._controller.state.ready.fetch()
        status = Status.STATUS_READY if ready else Status.STATUS_LOADING
        return GetControllerStatusResponse().from_dict({"status": status})

    async def watch_controller_status(
        self, watch_controller_status_request: "WatchControllerStatusRequest"
    ) -> AsyncIterator["WatchControllerStatusResponse"]:
        async for ready in self._controller.state.ready.subscribe():
            status = Status.STATUS_READY if ready else Status.STATUS_LOADING
            yield WatchControllerStatusResponse().from_dict({"status": status})

    async def get_face_status(
        self, get_face_status_request: "GetFaceStatusRequest"
    ) -> "GetFaceStatusResponse":
        status = await self._controller.get_face_status()
        status = Status(value=status.value)
        return GetFaceStatusResponse().from_dict({"status": status})

    async def watch_face_status(
        self, watch_face_status_request: "WatchFaceStatusRequest"
    ) -> AsyncIterator["WatchFaceStatusResponse"]:
        async for status in self._controller.watch_face_status():
            status = Status(value=status.value)
            yield WatchFaceStatusResponse().from_dict({"status": status})

    async def get_module_status(
        self, get_module_status_request: "GetModuleStatusRequest"
    ) -> "GetModuleStatusResponse":
        status = await self._controller.get_module_status()
        status = Status(value=status.value)
        return GetModuleStatusResponse().from_dict({"status": status})

    async def watch_module_status(
        self, watch_module_status_request: "WatchModuleStatusRequest"
    ) -> AsyncIterator["WatchModuleStatusResponse"]:
        async for status in self._controller.watch_module_status():
            status = Status(value=status.value)
            yield WatchModuleStatusResponse().from_dict({"status": status})

    async def get_training_status(
        self, get_training_status_request: "GetTrainingStatusRequest"
    ) -> "GetTrainingStatusResponse":
        status = await self._controller.get_training_status()
        status = await status.fetch()
        return GetTrainingStatusResponse(status=TrainingStatus(status.value))

    async def watch_training_status(
        self, watch_training_status_request: "WatchTrainingStatusRequest"
    ) -> AsyncIterator["WatchTrainingStatusResponse"]:
        status = await self._controller.get_training_status()
        async for status in status.subscribe():
            yield WatchTrainingStatusResponse(
                status=TrainingStatus(status.value)
            )

    async def get_controller_config_schema(
        self,
        get_controller_config_schema_request: "GetControllerConfigSchemaRequest",
    ) -> "GetControllerConfigSchemaResponse":
        schema = self._controller.schema
        return GetControllerConfigSchemaResponse().from_dict(
            {"schema": schema.json()}
        )

    async def get_controller_config(
        self, get_controller_config_request: "GetControllerConfigRequest"
    ) -> "GetControllerConfigResponse":
        config = await self._controller.config.json.fetch()
        return GetControllerConfigResponse().from_dict(
            {"config": json.dumps(config)}
        )

    async def watch_controller_config(
        self, watch_controller_config_request: "WatchControllerConfigRequest"
    ) -> AsyncIterator["WatchControllerConfigResponse"]:
        async for config in self._controller.config.json.subscribe():
            yield WatchControllerConfigResponse().from_dict(
                {"config": json.dumps(config)}
            )

    async def set_controller_config(
        self, set_controller_config_request: "SetControllerConfigRequest"
    ) -> "SetControllerConfigResponse":
        config = json.loads(set_controller_config_request.config)
        config = await self._controller.config.set(config)
        return SetControllerConfigResponse().from_dict(
            {"config": json.dumps(config)}
        )

    async def get_face_config_schema(
        self, get_face_config_schema_request: "GetFaceConfigSchemaRequest"
    ) -> "GetFaceConfigSchemaResponse":
        schema = await self._controller.get_face_config_schema()
        return GetFaceConfigSchemaResponse().from_dict(
            {"schema": schema.json()}
        )

    async def get_face_config(
        self, get_face_config_request: "GetFaceConfigRequest"
    ) -> "GetFaceConfigResponse":
        config = await self._controller.get_face_config()
        return GetFaceConfigResponse().from_dict(
            {"config": json.dumps(config)}
        )

    async def watch_face_config(
        self, watch_face_config_request: "WatchFaceConfigRequest"
    ) -> AsyncIterator["WatchFaceConfigResponse"]:
        async for config in self._controller.watch_face_config():
            yield WatchFaceConfigResponse().from_dict(
                {"config": json.dumps(config)}
            )

    async def set_face_config(
        self, set_face_config_request: "SetFaceConfigRequest"
    ) -> "SetFaceConfigResponse":
        config = json.loads(set_face_config_request.config)
        config = await self._controller.set_face_config(config)
        return SetFaceConfigResponse().from_dict(
            {"config": json.dumps(config)}
        )

    async def get_module_config_schema(
        self, get_module_config_schema_request: "GetModuleConfigSchemaRequest"
    ) -> "GetModuleConfigSchemaResponse":
        schema = await self._controller.get_module_config_schema()
        return GetModuleConfigSchemaResponse().from_dict(
            {"schema": schema.json()}
        )

    async def get_module_config(
        self, get_module_config_request: "GetModuleConfigRequest"
    ) -> "GetModuleConfigResponse":
        config = await self._controller.get_module_config()
        return GetModuleConfigResponse().from_dict(
            {"config": json.dumps(config)}
        )

    async def watch_module_config(
        self, watch_module_config_request: "WatchModuleConfigRequest"
    ) -> AsyncIterator["WatchModuleConfigResponse"]:
        async for config in self._controller.watch_module_config():
            yield WatchModuleConfigResponse().from_dict(
                {"config": json.dumps(config)}
            )

    async def set_module_config(
        self, set_module_config_request: "SetModuleConfigRequest"
    ) -> "SetModuleConfigResponse":
        config = json.loads(set_module_config_request.config)
        config = await self._controller.set_module_config(config)
        return SetModuleConfigResponse().from_dict(
            {"config": json.dumps(config)}
        )

    async def train_offline(
        self, train_offline_request: "TrainOfflineRequest"
    ) -> "TrainOfflineResponse":
        await self._controller.train_offline()
        return TrainOfflineResponse()

    async def train_online(
        self, train_online_request: "TrainOnlineRequest"
    ) -> "TrainOnlineResponse":
        await self._controller.train_online()
        return TrainOnlineResponse()

    async def stop_training(
        self, pause_training_request: "StopTrainingRequest"
    ) -> "StopTrainingResponse":
        await self._controller.stop_training()
        return StopTrainingResponse()

    async def get_module_metrics_config(
        self,
        get_module_metrics_config_request: "GetModuleMetricsConfigRequest",
    ) -> "GetModuleMetricsConfigResponse":
        metrics = await self._controller.get_module_metrics_config()
        return GetModuleMetricsConfigResponse(
            configs=[
                MetricConfig(
                    id=metric.id,
                    label=metric.label,
                    group=metric.group,
                    config=json.dumps(metric.config),
                )
                for metric in metrics
            ]
        )

    async def get_module_metrics(
        self, get_module_metrics_request: "GetModuleMetricsRequest"
    ) -> "GetModuleMetricsResponse":
        metrics = await self._controller.get_module_metrics()
        return GetModuleMetricsResponse(
            metrics=[
                MetricData(
                    metric_id=metric.metric_id,
                    dataset_id=metric.dataset_id,
                    data=json.dumps(metric.data),
                )
                for metric in metrics
            ]
        )

    async def watch_module_metrics(
        self, watch_module_metrics_request: "WatchModuleMetricsRequest"
    ) -> AsyncIterator["WatchModuleMetricsResponse"]:
        async for metric in self._controller.watch_module_metrics():
            yield WatchModuleMetricsResponse(
                metric=MetricData(
                    metric_id=metric.metric_id,
                    dataset_id=metric.dataset_id,
                    data=json.dumps(metric.data),
                )
            )

    async def watch_all(
        self, watch_all_request: "WatchAllRequest"
    ) -> AsyncIterator["WatchAllResponse"]:
        async def convert(
            method: str, stream: AsyncIterator[betterproto.Message]
        ) -> AsyncIterator[Tuple[str, str]]:
            async for msg in stream:
                yield method, msg.to_json()

        streams = [
            (
                "WatchControllerStatus",
                self.watch_controller_status(WatchControllerStatusRequest()),
            ),
            (
                "WatchFaceStatus",
                self.watch_face_status(WatchFaceStatusRequest()),
            ),
            (
                "WatchModuleStatus",
                self.watch_module_status(WatchModuleStatusRequest()),
            ),
            (
                "WatchTrainingStatus",
                self.watch_training_status(WatchTrainingStatusRequest()),
            ),
            (
                "WatchControllerConfig",
                self.watch_controller_config(WatchControllerConfigRequest()),
            ),
            (
                "WatchFaceConfig",
                self.watch_face_config(WatchFaceConfigRequest()),
            ),
            (
                "WatchModuleConfig",
                self.watch_module_config(WatchModuleConfigRequest()),
            ),
            (
                "WatchModuleMetrics",
                self.watch_module_metrics(WatchModuleMetricsRequest()),
            ),
            (
                "WatchFeed",
                self.watch_feed(WatchFeedRequest()),
            ),
        ]

        combine = aiostream.stream.merge(
            *(convert(method, stream) for method, stream in streams)
        )

        async with combine.stream() as streamer:
            async for method, message in streamer:
                yield WatchAllResponse(method=method, message=message)

    async def reset_controller(
        self, reset_controller_request: "ResetControllerRequest"
    ) -> "ResetControllerResponse":
        await self._controller.init()
        return ResetControllerResponse()

    async def reset_face(
        self, reset_face_request: "ResetFaceRequest"
    ) -> "ResetFaceResponse":
        await self._controller.reset_face()
        return ResetFaceResponse()

    async def reset_module(
        self, reset_module_request: "ResetModuleRequest"
    ) -> "ResetModuleResponse":
        await self._controller.reset_module()
        return ResetModuleResponse()

    async def get_feed(
        self, get_feed_request: "GetFeedRequest"
    ) -> "GetFeedResponse":
        feed = await self._controller.get_feed()
        return GetFeedResponse(
            posts=[
                Post(
                    id=str(post.id),
                    url=post.url,
                    content=json.dumps(post.content),
                    created_at=post.created_at.isoformat().replace(
                        "+00:00", "Z"
                    ),
                )
                for post in feed
            ]
        )

    async def watch_feed(
        self, watch_feed_request: "WatchFeedRequest"
    ) -> AsyncIterator["WatchFeedResponse"]:
        async for post in self._controller.watch_feed():
            yield WatchFeedResponse(
                post=Post(
                    id=str(post.id),
                    url=post.url,
                    content=json.dumps(post.content),
                    created_at=post.created_at.isoformat().replace(
                        "+00:00", "Z"
                    ),
                )
            )



from datetime import datetime
from typing import Any, Awaitable, Callable, Dict, Optional

from kilroy_server_py_utils import (
    CategorizableBasedParameter,
    OptionalParameter,
    Parameter,
    classproperty,
)

from cilroy.controller.state import State
from cilroy.posting import PostScheduler
from cilroy.scoring import ScoreScheduler


class ScrapBeforeParameter(OptionalParameter[State, str]):
    async def _get(self, state: State) -> Optional[str]:
        if state.offline.scrap_before is None:
            return None
        return state.offline.scrap_before.isoformat()

    async def _set(
        self, state: State, value: Optional[str]
    ) -> Callable[[], Awaitable]:
        original_value = state.offline.scrap_before

        async def undo():
            state.offline.scrap_before = original_value

        state.offline.scrap_before = (
            datetime.fromisoformat(value) if value is not None else None
        )
        return undo

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": ["string", "null"],
            "format": "date-time",
            "title": cls.pretty_name,
            "default": None,
        }


class ScrapAfterParameter(OptionalParameter[State, str]):
    async def _get(self, state: State) -> Optional[str]:
        if state.offline.scrap_after is None:
            return None
        return state.offline.scrap_after.isoformat()

    async def _set(
        self, state: State, value: Optional[str]
    ) -> Callable[[], Awaitable]:
        original_value = state.offline.scrap_after

        async def undo():
            state.offline.scrap_after = original_value

        state.offline.scrap_after = (
            datetime.fromisoformat(value) if value is not None else None
        )
        return undo

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": ["string", "null"],
            "format": "date-time",
            "title": cls.pretty_name,
            "default": None,
        }


class ScrapLimitParameter(OptionalParameter[State, int]):
    async def _get(self, state: State) -> Optional[int]:
        if state.offline.scrap_limit is None:
            return None
        return state.offline.scrap_limit

    async def _set(
        self, state: State, value: Optional[int]
    ) -> Callable[[], Awaitable]:
        original_value = state.offline.scrap_limit

        async def undo():
            state.offline.scrap_limit = original_value

        state.offline.scrap_limit = value if value is not None else None
        return undo

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": ["integer", "null"],
            "minimum": 0,
            "title": cls.pretty_name,
            "default": None,
        }


class MaxOfflineEpochsParameter(OptionalParameter[State, int]):
    async def _get(self, state: State) -> Optional[int]:
        if state.offline.max_epochs is None:
            return None
        return state.offline.max_epochs

    async def _set(
        self, state: State, value: Optional[int]
    ) -> Callable[[], Awaitable]:
        original_value = state.offline.max_epochs

        async def undo():
            state.offline.max_epochs = original_value

        state.offline.max_epochs = value if value is not None else None
        return undo

    @classproperty
    def pretty_name(cls) -> str:
        return "Maximum Offline Epochs"

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": ["integer", "null"],
            "minimum": 0,
            "title": cls.pretty_name,
            "default": None,
        }


class OfflineBatchSizeParameter(OptionalParameter[State, int]):
    async def _get(self, state: State) -> Optional[int]:
        if state.offline.batch_size is None:
            return None
        return state.offline.batch_size

    async def _set(
        self, state: State, value: Optional[int]
    ) -> Callable[[], Awaitable]:
        original_value = state.offline.batch_size

        async def undo():
            state.offline.batch_size = original_value

        state.offline.batch_size = value if value is not None else None
        return undo

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": ["integer", "null"],
            "minimum": 1,
            "title": cls.pretty_name,
            "default": None,
        }


class PostSchedulerParameter(
    CategorizableBasedParameter[State, PostScheduler]
):
    async def _get_categorizable(self, state: State) -> PostScheduler:
        return state.online.post_scheduler

    async def _set_categorizable(
        self, state: State, value: PostScheduler
    ) -> None:
        state.online.post_scheduler = value

    async def _get_params(self, state: State, category: str) -> Dict[str, Any]:
        return state.online.post_schedulers_params[category]


class ScoreSchedulerParameter(
    CategorizableBasedParameter[State, ScoreScheduler]
):
    async def _get_categorizable(self, state: State) -> ScoreScheduler:
        return state.online.score_scheduler

    async def _set_categorizable(
        self, state: State, value: ScoreScheduler
    ) -> None:
        state.online.score_scheduler = value

    async def _get_params(self, state: State, category: str) -> Dict[str, Any]:
        return state.online.score_schedulers_params[category]


class OnlineIterationsParameter(Parameter[State, int]):
    async def _get(self, state: State) -> int:
        return state.online.iterations

    async def _set(self, state: State, value: int) -> Callable[[], Awaitable]:
        original_value = state.online.iterations

        async def undo():
            state.online.iterations = original_value

        state.online.iterations = value
        return undo

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": "integer",
            "minimum": 0,
            "title": cls.pretty_name,
            "default": 1,
        }


class OnlineBatchSizeParameter(OptionalParameter[State, int]):
    async def _get(self, state: State) -> Optional[int]:
        if state.online.batch_size is None:
            return None
        return state.online.batch_size

    async def _set(
        self, state: State, value: Optional[int]
    ) -> Callable[[], Awaitable]:
        original_value = state.online.batch_size

        async def undo():
            state.online.batch_size = original_value

        state.online.batch_size = value if value is not None else None
        return undo

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": ["integer", "null"],
            "minimum": 1,
            "title": cls.pretty_name,
            "default": None,
        }


class FeedLengthParameter(Parameter[State, int]):
    async def _get(self, state: State) -> int:
        return state.feed.length

    async def _set(self, state: State, value: int) -> Callable[[], Awaitable]:
        original_value = state.feed.length
        original_feed = state.feed.feed

        async def undo():
            state.feed.length = original_value
            state.feed.feed = original_feed

        state.feed.length = value
        state.feed.feed = state.feed.feed[-value:]
        return undo

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": "integer",
            "minimum": 1,
            "title": cls.pretty_name,
            "default": 100,
        }



"""Main script.

This module provides basic CLI entrypoint.

"""
import asyncio
import logging
import signal
from asyncio import FIRST_EXCEPTION
from pathlib import Path
from typing import List, Optional

import typer
from typer import FileText

from cilroy import log
from cilroy.config import Config, get_config
from cilroy.controller import CilroyController
from cilroy.server import CilroyServer

cli = typer.Typer()  # this is actually callable and thus can be an entry point

logger = logging.getLogger(__name__)


async def shutdown(sig: signal.Signals, loop: asyncio.AbstractEventLoop):
    loop.remove_signal_handler(sig)
    logger.info(f"Received exit signal {sig.name}...")
    tasks = [
        task
        for task in asyncio.all_tasks()
        if task is not asyncio.current_task()
    ]
    logger.info(f"Cancelling {len(tasks)} outstanding tasks...")
    for task in tasks:
        task.cancel()
    await asyncio.gather(*tasks, return_exceptions=True)
    loop.stop()


async def attach_signal_handlers() -> None:
    loop = asyncio.get_running_loop()
    for sig in (signal.SIGINT, signal.SIGTERM):
        loop.add_signal_handler(
            sig, lambda s=sig: asyncio.create_task(shutdown(s, loop))
        )


async def load_or_init(controller: CilroyController, state_dir: Path) -> None:
    if not state_dir.exists() or not any(state_dir.iterdir()):
        logger.info("Initializing controller...")
        await controller.init()
        logger.info("Initialization complete.")
        return

    try:
        logger.info("Loading state...")
        await controller.load_saved(state_dir)
        logger.info("Loading complete.")
    except Exception as e:
        logger.warning(
            "Can't load saved state. Will try to initialize instead.",
            exc_info=e,
        )
        logger.info("Initializing controller...")
        await controller.init()
        logger.info("Initialization complete.")


async def run(config: Config) -> None:
    await attach_signal_handlers()

    face = await CilroyController.build(**config.controller.dict())
    server = CilroyServer(face, logger)

    server_task = asyncio.create_task(server.run(**config.server.dict()))
    init_task = asyncio.create_task(load_or_init(face, config.state_directory))

    tasks = [server_task, init_task]

    try:
        done, pending = await asyncio.wait(tasks, return_when=FIRST_EXCEPTION)
    except asyncio.CancelledError:
        done, pending = [], tasks
    except Exception as e:
        logger.warning("Unhandled exception.", exc_info=e)
        done, pending = [], tasks

    for task in pending:
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass
        except Exception as e:
            logger.warning("Unhandled exception.", exc_info=e)

    for task in done:
        task.result()

    if (
        init_task.done()
        and not init_task.cancelled()
        and init_task.exception() is None
    ):
        logger.info("Saving state...")
        await face.save(config.state_directory)

        logger.info("Cleaning up...")
        await face.cleanup()


@cli.command()
def main(
    config_file: Optional[FileText] = typer.Option(
        None, "--config-file", "-C", dir_okay=False, help="Configuration file."
    ),
    config: Optional[List[str]] = typer.Option(
        None, "--config", "-c", help="Configuration entries."
    ),
    verbosity: log.Verbosity = typer.Option(
        "INFO", "--verbosity", "-v", help="Verbosity level."
    ),
) -> None:
    """Command line interface for cilroy."""

    log.configure(verbosity)

    logger.info("Loading config...")
    try:
        config = get_config(config_file, config)
    except ValueError as e:
        logger.error("Failed to parse config!", exc_info=e)
        raise typer.Exit(1)
    logger.info("Config loaded!")

    asyncio.run(run(config))


if __name__ == "__main__":
    # entry point for "python -m"
    cli()



from importlib.resources import path
from pathlib import Path, PurePath
from typing import ContextManager, Tuple, Union


def extract_resource_path(
    resource_path: Union[str, PurePath]
) -> Tuple[str, str]:
    """Returns path to resource as tuple with parent part (as package with dots) and resource name."""
    resource_path = PurePath(resource_path)
    if not resource_path.is_relative_to(PurePath(".")):
        raise ValueError(
            f"Path has to be relative to current dir, but '{resource_path}' is not"
        )
    if ".." in resource_path.parts:
        raise ValueError(
            f"Path can't go backwards, but '{resource_path}' does"
        )
    return ".".join(resource_path.parent.parts), resource_path.name


def resource(resource_path: Union[str, PurePath]) -> ContextManager[Path]:
    """Wrapper to importlib.resources.path with package part filled."""
    package_part, resource_name = extract_resource_path(resource_path)
    package_part = f"{__name__}{'.' + package_part if package_part else ''}"
    return path(package_part, resource_name)


def resource_bytes(resource_path: Union[str, PurePath]) -> bytes:
    """Wrapper to Path.read_bytes() that uses resource(resource_path)."""
    with resource(resource_path) as r:
        return r.read_bytes()


def resource_text(
    resource_path: Union[str, PurePath],
    encoding: str = None,
    errors: str = None,
) -> str:
    """Wrapper to Path.read_text() that uses resource(resource_path)."""
    with resource(resource_path) as r:
        return r.read_text(encoding, errors)



import asyncio
import json
import logging
from abc import ABC, abstractmethod
from asyncio import Lock
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Dict,
    List,
    Mapping,
    Set,
    Tuple,
)
from uuid import UUID

from aiostream.aiter_utils import anext
from grpclib.client import Channel
from kilroy_face_client_py_sdk import FaceService
from kilroy_module_client_py_sdk import MetricConfig, MetricData, ModuleService
from kilroy_server_py_utils import (
    Configurable,
    JSONSchema,
    Observable,
    Parameter,
    ReadOnlyObservableWrapper,
    ReadableObservable,
    classproperty,
)

from cilroy.controller.parameters import (
    MaxOfflineEpochsParameter,
    OfflineBatchSizeParameter,
    OnlineBatchSizeParameter,
    OnlineIterationsParameter,
    PostSchedulerParameter,
    ScoreSchedulerParameter,
    ScrapAfterParameter,
    ScrapBeforeParameter,
    ScrapLimitParameter,
    FeedLengthParameter,
)
from cilroy.controller.params import (
    OfflineParams,
    OnlineParams,
    Params,
    FaceParams,
    ModuleParams,
    FeedParams,
)
from cilroy.controller.state import (
    OfflineState,
    OnlineState,
    State,
    FaceState,
    ModuleState,
    FeedState,
)
from cilroy.messages import Status
from cilroy.metadata import Metadata
from cilroy.post import PostData
from cilroy.posting import PostScheduler
from cilroy.scoring import ScoreScheduler
from cilroy.status import TrainingStatus
from cilroy.utils import CachingAsyncIterable, batches

logger = logging.getLogger(__name__)


class CilroyControllerBase(Configurable[State], ABC):
    @staticmethod
    async def _build_face_service(params: FaceParams) -> FaceService:
        return FaceService(Channel(params.host, params.port))

    @classmethod
    async def _build_face_state(cls, params: FaceParams) -> FaceState:
        return FaceState(
            service=await cls._build_face_service(params),
        )

    @staticmethod
    async def _build_module_service(params: ModuleParams) -> ModuleService:
        return ModuleService(Channel(params.host, params.port))

    async def _build_module_state(self, params: ModuleParams) -> ModuleState:
        return ModuleState(
            service=await self._build_module_service(params),
            archived_metrics=[],
            metrics_stream=await Observable.build(),
            metrics_task=asyncio.create_task(self._watch_metrics_task()),
        )

    @staticmethod
    async def _build_offline_state(params: OfflineParams) -> OfflineState:
        return OfflineState(
            max_epochs=params.max_epochs,
            batch_size=params.batch_size,
            scrap_limit=params.scrap_limit,
            scrap_before=params.scrap_before,
            scrap_after=params.scrap_after,
            posts_cache={},
        )

    @classmethod
    async def _build_post_scheduler(
        cls, params: OnlineParams
    ) -> PostScheduler:
        return await cls._build_categorizable(
            PostScheduler,
            params.post_scheduler_type,
            **params.post_schedulers_params.get(
                params.post_scheduler_type, {}
            ),
        )

    @classmethod
    async def _build_score_scheduler(
        cls, params: OnlineParams
    ) -> ScoreScheduler:
        return await cls._build_categorizable(
            ScoreScheduler,
            params.score_scheduler_type,
            **params.score_schedulers_params.get(
                params.score_scheduler_type, {}
            ),
        )

    @classmethod
    async def _build_online_state(cls, params: OnlineParams) -> OnlineState:
        return OnlineState(
            ids_cache={},
            post_scheduler=await cls._build_post_scheduler(params),
            post_schedulers_params=params.post_schedulers_params,
            score_scheduler=await cls._build_score_scheduler(params),
            score_schedulers_params=params.score_schedulers_params,
            iterations=params.iterations,
            batch_size=params.batch_size,
            lock=Lock(),
        )

    @staticmethod
    async def _build_training_status() -> Observable[TrainingStatus]:
        return await Observable.build(TrainingStatus.IDLE)

    @staticmethod
    async def _build_feed_state(params: FeedParams) -> FeedState:
        return FeedState(
            feed=[],
            length=params.length,
            stream=await Observable.build(),
        )

    async def _build_default_state(self) -> State:
        params = Params(**self._kwargs)
        return State(
            face=await self._build_face_state(params.face),
            module=await self._build_module_state(params.module),
            offline=await self._build_offline_state(params.offline),
            online=await self._build_online_state(params.online),
            training_task=None,
            training_status=await self._build_training_status(),
            feed=await self._build_feed_state(params.feed),
        )

    @staticmethod
    async def _save_state_dict(
        directory: Path, state_dict: Dict[str, Any]
    ) -> None:
        with open(directory / "state.json", "w") as f:
            json.dump(state_dict, f)

    @staticmethod
    async def _save_face_state(state: FaceState, directory: Path) -> None:
        pass

    @staticmethod
    async def _save_module_metrics(
        metrics: List[MetricData], directory: Path
    ) -> None:
        serialized = [json.loads(metric.json()) for metric in metrics]
        with open(directory / "metrics.json", "w") as f:
            json.dump([metric for metric in serialized], f)

    @classmethod
    async def _save_module_state(
        cls, state: ModuleState, directory: Path
    ) -> None:
        metrics_directory = directory / "metrics"
        metrics_directory.mkdir(parents=True, exist_ok=True)
        await cls._save_module_metrics(
            state.archived_metrics, metrics_directory
        )

    @staticmethod
    async def _create_offline_state_dict(
        state: OfflineState,
    ) -> Dict[str, Any]:
        return {
            "max_epochs": state.max_epochs,
            "batch_size": state.batch_size,
            "scrap_limit": state.scrap_limit,
            "scrap_before": state.scrap_before.isoformat()
            if state.scrap_before is not None
            else None,
            "scrap_after": state.scrap_after.isoformat()
            if state.scrap_after is not None
            else None,
        }

    @classmethod
    async def _save_offline_state(
        cls, state: OfflineState, directory: Path
    ) -> None:
        state_dict = await cls._create_offline_state_dict(state)
        await cls._save_state_dict(directory, state_dict)

    @staticmethod
    async def _serialize_ids_cache(
        ids_cache: Mapping[UUID, UUID]
    ) -> Dict[str, str]:
        return {str(k): str(v) for k, v in ids_cache.items()}

    @staticmethod
    async def _deserialize_ids_cache(
        ids_cache: Mapping[str, str]
    ) -> Dict[UUID, UUID]:
        return {UUID(k): UUID(v) for k, v in ids_cache.items()}

    @classmethod
    async def _create_online_state_dict(
        cls,
        state: OnlineState,
    ) -> Dict[str, Any]:
        return {
            "ids_cache": await cls._serialize_ids_cache(state.ids_cache),
            "post_scheduler_type": state.post_scheduler.category,
            "post_schedulers_params": state.post_schedulers_params,
            "score_scheduler_type": state.score_scheduler.category,
            "score_schedulers_params": state.score_schedulers_params,
            "iterations": state.iterations,
            "batch_size": state.batch_size,
        }

    @staticmethod
    async def _save_post_scheduler(
        scheduler: PostScheduler, directory: Path
    ) -> None:
        if isinstance(scheduler, Configurable):
            await scheduler.save(directory)

    @staticmethod
    async def _save_score_scheduler(
        scheduler: ScoreScheduler, directory: Path
    ) -> None:
        if isinstance(scheduler, Configurable):
            await scheduler.save(directory)

    @classmethod
    async def _save_online_state(
        cls, state: OnlineState, directory: Path
    ) -> None:
        post_scheduler_dir = directory / "post_scheduler"
        post_scheduler_dir.mkdir(parents=True, exist_ok=True)
        await cls._save_post_scheduler(
            state.post_scheduler, post_scheduler_dir
        )

        score_scheduler_dir = directory / "score_scheduler"
        score_scheduler_dir.mkdir(parents=True, exist_ok=True)
        await cls._save_score_scheduler(
            state.score_scheduler, score_scheduler_dir
        )

        state_dict = await cls._create_online_state_dict(state)
        await cls._save_state_dict(directory, state_dict)

    @classmethod
    async def _save_feed_state(cls, state: FeedState, directory: Path) -> None:
        serialized_feed = [json.loads(post.json()) for post in state.feed]
        state_dict = {
            "feed": serialized_feed,
            "length": state.length,
        }
        await cls._save_state_dict(directory, state_dict)

    @classmethod
    async def _save_state(cls, state: State, directory: Path) -> None:
        face_directory = directory / "face"
        face_directory.mkdir(parents=True, exist_ok=True)
        await cls._save_face_state(state.face, face_directory)

        module_directory = directory / "module"
        module_directory.mkdir(parents=True, exist_ok=True)
        await cls._save_module_state(state.module, module_directory)

        offline_directory = directory / "offline"
        offline_directory.mkdir(parents=True, exist_ok=True)
        await cls._save_offline_state(state.offline, offline_directory)

        online_directory = directory / "online"
        online_directory.mkdir(parents=True, exist_ok=True)
        await cls._save_online_state(state.online, online_directory)

        feed_directory = directory / "feed"
        feed_directory.mkdir(parents=True, exist_ok=True)
        await cls._save_feed_state(state.feed, feed_directory)

    @staticmethod
    async def _load_state_dict(directory: Path) -> Dict[str, Any]:
        with open(directory / "state.json", "r") as f:
            return json.load(f)

    @classmethod
    async def _load_saved_face_state(
        cls, directory: Path, params: FaceParams
    ) -> FaceState:
        return FaceState(service=await cls._build_face_service(params))

    @staticmethod
    async def _load_module_metrics(directory: Path) -> List[MetricData]:
        with open(directory / "metrics.json", "r") as f:
            serialized = json.load(f)
        return [MetricData.parse_obj(metric) for metric in serialized]

    async def _load_saved_module_state(
        self, directory: Path, params: ModuleParams
    ) -> ModuleState:
        return ModuleState(
            service=await self._build_module_service(params),
            archived_metrics=await self._load_module_metrics(
                directory / "metrics"
            ),
            metrics_stream=await Observable.build(),
            metrics_task=asyncio.create_task(self._watch_metrics_task()),
        )

    @classmethod
    async def _load_saved_offline_state(
        cls, directory: Path, params: OfflineParams
    ) -> OfflineState:
        state_dict = await cls._load_state_dict(directory)
        return OfflineState(
            max_epochs=state_dict["max_epochs"],
            batch_size=state_dict["batch_size"],
            scrap_limit=state_dict["scrap_limit"],
            scrap_before=datetime.fromisoformat(state_dict["scrap_before"])
            if state_dict["scrap_before"] is not None
            else None,
            scrap_after=datetime.fromisoformat(state_dict["scrap_after"])
            if state_dict["scrap_after"] is not None
            else None,
            posts_cache={},
        )

    @classmethod
    async def _load_saved_post_scheduler(
        cls,
        directory: Path,
        state_dict: Dict[str, Any],
        params: OnlineParams,
    ) -> PostScheduler:
        return await cls._load_generic(
            directory,
            PostScheduler,
            category=state_dict["post_scheduler_type"],
            default=partial(
                cls._build_post_scheduler,
                params,
            ),
        )

    @classmethod
    async def _load_saved_score_scheduler(
        cls,
        directory: Path,
        state_dict: Dict[str, Any],
        params: OnlineParams,
    ) -> ScoreScheduler:
        return await cls._load_generic(
            directory,
            ScoreScheduler,
            category=state_dict["score_scheduler_type"],
            default=partial(
                cls._build_score_scheduler,
                params,
            ),
        )

    @classmethod
    async def _load_saved_online_state(
        cls, directory: Path, params: OnlineParams
    ) -> OnlineState:
        state_dict = await cls._load_state_dict(directory)
        return OnlineState(
            ids_cache=await cls._deserialize_ids_cache(
                state_dict["ids_cache"]
            ),
            post_scheduler=await cls._load_saved_post_scheduler(
                directory / "post_scheduler",
                state_dict,
                params,
            ),
            post_schedulers_params=state_dict["post_schedulers_params"],
            score_scheduler=await cls._load_saved_score_scheduler(
                directory / "score_scheduler",
                state_dict,
                params,
            ),
            score_schedulers_params=state_dict["score_schedulers_params"],
            iterations=state_dict["iterations"],
            batch_size=state_dict["batch_size"],
            lock=Lock(),
        )

    @classmethod
    async def _load_saved_feed_state(
        cls, directory: Path, params: FeedParams
    ) -> FeedState:
        state_dict = await cls._load_state_dict(directory)
        return FeedState(
            feed=[
                PostData.parse_obj(post) for post in state_dict.get("feed", [])
            ],
            length=state_dict.get("length", params.length),
            stream=await Observable.build(),
        )

    async def _load_saved_state(self, directory: Path) -> State:
        params = Params(**self._kwargs)
        return State(
            face=await self._load_saved_face_state(
                directory / "face", params.face
            ),
            module=await self._load_saved_module_state(
                directory / "module", params.module
            ),
            offline=await self._load_saved_offline_state(
                directory / "offline", params.offline
            ),
            online=await self._load_saved_online_state(
                directory / "online", params.online
            ),
            training_task=None,
            training_status=await self._build_training_status(),
            feed=await self._load_saved_feed_state(
                directory / "feed", params.feed
            ),
        )

    async def cleanup(self) -> None:
        async with self.state.write_lock() as state:
            state.module.metrics_task.cancel()
            try:
                await state.module.metrics_task
            except asyncio.CancelledError:
                pass

    @classproperty
    def parameters(cls) -> Set[Parameter]:
        return {
            ScrapBeforeParameter(),
            ScrapAfterParameter(),
            ScrapLimitParameter(),
            MaxOfflineEpochsParameter(),
            OfflineBatchSizeParameter(),
            PostSchedulerParameter(),
            ScoreSchedulerParameter(),
            OnlineIterationsParameter(),
            OnlineBatchSizeParameter(),
            FeedLengthParameter(),
        }

    @abstractmethod
    async def _watch_metrics_task(self) -> None:
        pass


class CilroyControllerDelegatedBase(CilroyControllerBase, ABC):
    async def get_face_metadata(self) -> Metadata:
        async with self.state.read_lock() as state:
            metadata = await state.face.service.get_metadata()
            return Metadata(key=metadata.key, description=metadata.description)

    async def get_module_metadata(self) -> Metadata:
        async with self.state.read_lock() as state:
            metadata = await state.module.service.get_metadata()
            return Metadata(key=metadata.key, description=metadata.description)

    async def get_face_post_schema(self) -> JSONSchema:
        async with self.state.read_lock() as state:
            return JSONSchema(**await state.face.service.get_post_schema())

    async def get_module_post_schema(self) -> JSONSchema:
        async with self.state.read_lock() as state:
            return JSONSchema(**await state.module.service.get_post_schema())

    async def get_face_status(self) -> Status:
        async with self.state.read_lock() as state:
            status = await state.face.service.get_status()
        return Status(status)

    async def watch_face_status(self) -> AsyncIterable[Status]:
        async with self.state.read_lock() as state:
            service = state.face.service
        async for status in service.watch_status():
            yield Status(status)

    async def get_module_status(self) -> Status:
        async with self.state.read_lock() as state:
            status = await state.module.service.get_status()
        return Status(status)

    async def watch_module_status(self) -> AsyncIterable[Status]:
        async with self.state.read_lock() as state:
            service = state.module.service
        async for status in service.watch_status():
            yield Status(status)

    async def get_face_config_schema(self) -> JSONSchema:
        async with self.state.read_lock() as state:
            return JSONSchema(**await state.face.service.get_config_schema())

    async def get_face_config(self) -> Dict[str, Any]:
        async with self.state.read_lock() as state:
            return await state.face.service.get_config()

    async def watch_face_config(self) -> AsyncIterable[Dict[str, Any]]:
        async with self.state.read_lock() as state:
            service = state.face.service
        async for config in service.watch_config():
            yield config

    async def set_face_config(self, config: Dict[str, Any]) -> Dict[str, Any]:
        async with self.state.write_lock() as state:
            return await state.face.service.set_config(config)

    async def get_module_config_schema(self) -> JSONSchema:
        async with self.state.read_lock() as state:
            return JSONSchema(**await state.module.service.get_config_schema())

    async def get_module_config(self) -> Dict[str, Any]:
        async with self.state.read_lock() as state:
            return await state.module.service.get_config()

    async def watch_module_config(self) -> AsyncIterable[Dict[str, Any]]:
        async with self.state.read_lock() as state:
            service = state.module.service
        async for config in service.watch_config():
            yield config

    async def set_module_config(
        self, config: Dict[str, Any]
    ) -> Dict[str, Any]:
        async with self.state.write_lock() as state:
            return await state.module.service.set_config(config)


class CilroyController(CilroyControllerDelegatedBase):
    async def get_training_status(self) -> ReadableObservable[TrainingStatus]:
        async with self.state.read_lock() as state:
            return ReadOnlyObservableWrapper(state.training_status)

    async def _train_offline_loop(
        self, posts: AsyncIterable[Tuple[Dict, float]]
    ) -> None:
        epoch = 0

        while True:
            logger.info(f"Offline training: Epoch {epoch}.")

            async with self.state.read_lock() as state:
                max_epochs = state.offline.max_epochs
                batch_size = state.offline.batch_size
                service = state.module.service

            if max_epochs is not None and epoch >= max_epochs:
                break

            n_batch = 0

            async for batch in batches(posts, batch_size):
                logger.info(
                    f"Offline training: Epoch {epoch}, batch {n_batch}."
                )
                await service.fit_posts(batch)
                await service.step()
                n_batch += 1

            epoch += 1

    async def _train_offline(self) -> None:
        logger.info("Offline training started.")

        async def map_posts(
            _posts: AsyncIterable[Tuple[UUID, Dict, float]]
        ) -> AsyncIterable[Tuple[Dict, float]]:
            async for post_id, post, score in _posts:
                yield post, score

        state = await self.state.value.fetch()
        posts = state.face.service.scrap(
            state.offline.scrap_limit,
            state.offline.scrap_before,
            state.offline.scrap_after,
        )
        posts = map_posts(posts)

        with CachingAsyncIterable(posts, state.offline.posts_cache) as posts:
            await self._train_offline_loop(posts)

        async with self.state.write_lock() as state:
            await state.training_status.set(TrainingStatus.IDLE)
            state.training_task = None

        logger.info("Offline training finished.")

    async def train_offline(self) -> None:
        async with self.state.write_lock() as state:
            if state.training_task is not None:
                raise RuntimeError("Training is already in progress.")
            await state.training_status.set(TrainingStatus.OFFLINE)
            state.training_task = asyncio.create_task(self._train_offline())

    async def _train_online_post_loop(self) -> None:
        async with self.state.read_lock() as state:
            module_service = state.module.service
            face_service = state.face.service
            post_scheduler = state.online.post_scheduler
            lock = state.online.lock

        async for _ in post_scheduler.wait():
            async with lock:
                logger.info("Online training: Creating new post...")

                module_post_id, post = await anext(module_service.generate())
                face_post_id, post_url = await face_service.post(post)

                post_data = PostData(
                    id=face_post_id,
                    url=post_url,
                    content=post,
                    created_at=datetime.utcnow(),
                )

                async with self.state.write_lock() as state:
                    state.online.ids_cache[face_post_id] = module_post_id
                    state.feed.feed = state.feed.feed + [post_data]
                    state.feed.feed = state.feed.feed[-state.feed.length :]
                    await state.feed.stream.set(post_data)

                logger.info(
                    f"Online training: "
                    f"New post with module-side id {str(module_post_id)} and "
                    f"face-side id {str(face_post_id)}."
                )

    async def _train_online_fit_scores(
        self, scores: Dict[UUID, float]
    ) -> None:
        iteration = 0

        while True:
            async with self.state.read_lock() as state:
                iterations = state.online.iterations
                batch_size = state.online.batch_size
                service = state.module.service

            if iteration >= iterations:
                break

            async for batch in batches(scores.items(), batch_size):
                await service.fit_scores(batch)
                await service.step()
            iteration += 1

    async def _train_online_score_loop(self) -> None:
        async with self.state.read_lock() as state:
            face_service = state.face.service
            score_scheduler = state.online.score_scheduler
            lock = state.online.lock

        async for _ in score_scheduler.wait():
            async with lock:
                logger.info("Online training: Scoring posts...")

                async with self.state.read_lock() as state:
                    post_ids = list(state.online.ids_cache.keys())

                scores = await asyncio.gather(
                    *[face_service.score(post_id) for post_id in post_ids]
                )

                async with self.state.read_lock() as state:
                    scores = {
                        state.online.ids_cache.get(post_id): score
                        for post_id, score in zip(post_ids, scores)
                    }

                logger.info("Online training: Fitting scores...")

                await self._train_online_fit_scores(scores)

                async with self.state.write_lock() as state:
                    state.online.ids_cache.clear()

                logger.info("Online training: Scores fitted.")

    async def _train_online(self) -> None:
        logger.info("Online training started.")

        tasks = [
            asyncio.create_task(self._train_online_post_loop()),
            asyncio.create_task(self._train_online_score_loop()),
        ]
        try:
            await asyncio.gather(*tasks)
        finally:
            for task in tasks:
                task.cancel()

        async with self.state.write_lock() as state:
            await state.training_status.set(TrainingStatus.IDLE)
            state.training_task = None

        logger.info("Online training finished.")

    async def train_online(self) -> None:
        async with self.state.write_lock() as state:
            if state.training_task is not None:
                raise RuntimeError("Training is already in progress.")
            await state.training_status.set(TrainingStatus.ONLINE)
            state.training_task = asyncio.create_task(self._train_online())

    async def stop_training(self) -> None:
        async with self.state.write_lock() as state:
            if state.training_task is not None:
                state.training_task.cancel()
                try:
                    await state.training_task
                except asyncio.CancelledError:
                    pass
                except Exception as e:
                    logger.warning("Training task failed.", exc_info=e)
                state.training_task = None
                await state.training_status.set(TrainingStatus.IDLE)
                logger.info("Training stopped.")

    async def get_module_metrics_config(self) -> List[MetricConfig]:
        async with self.state.read_lock() as state:
            return await state.module.service.get_metrics_config()

    async def get_module_metrics(self) -> List[MetricData]:
        async with self.state.read_lock() as state:
            return state.module.archived_metrics

    async def _watch_metrics_task(self) -> None:
        async with self.state.read_lock() as state:
            service: ModuleService = state.module.service

        while True:
            try:
                async for metric in service.watch_metrics():
                    async with self.state.write_lock() as state:
                        state.module.archived_metrics.append(metric)
                        await state.module.metrics_stream.set(metric)
            except asyncio.CancelledError:
                raise
            except Exception as e:
                logger.warning("Metrics stream failed.", exc_info=e)
                await asyncio.sleep(1)

    async def watch_module_metrics(self) -> AsyncIterator[MetricData]:
        async with self.state.read_lock() as state:
            metrics = state.module.metrics_stream
        async for metric in metrics.subscribe():
            yield metric

    async def reset_face(self) -> None:
        async with self.state.write_lock() as state:
            await state.face.service.reset()

    async def reset_module(self) -> None:
        await self.stop_training()
        async with self.state.write_lock() as state:
            await state.module.service.reset()
            state.module.archived_metrics = []
            state.offline.posts_cache.clear()
            state.online.ids_cache.clear()
            state.feed.feed = []

    async def get_feed(self) -> List[PostData]:
        async with self.state.read_lock() as state:
            return state.feed.feed

    async def watch_feed(self) -> AsyncIterator[PostData]:
        async with self.state.read_lock() as state:
            feed = state.feed.stream
        async for post in feed.subscribe():
            yield post



import asyncio
from asyncio import Lock
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, MutableMapping, Optional, Tuple
from uuid import UUID

from kilroy_face_client_py_sdk import FaceService
from kilroy_module_client_py_sdk import MetricData, ModuleService
from kilroy_server_py_utils import Observable

from cilroy.post import PostData
from cilroy.posting import PostScheduler
from cilroy.scoring import ScoreScheduler
from cilroy.status import TrainingStatus


@dataclass
class FaceState:
    service: FaceService


@dataclass
class ModuleState:
    service: ModuleService
    archived_metrics: List[MetricData]
    metrics_stream: Observable[MetricData]
    metrics_task: asyncio.Task


@dataclass
class OfflineState:
    scrap_before: Optional[datetime]
    scrap_after: Optional[datetime]
    scrap_limit: Optional[int]
    max_epochs: Optional[int]
    batch_size: Optional[int]
    posts_cache: MutableMapping[str, Tuple[Dict, float]]


@dataclass
class OnlineState:
    ids_cache: MutableMapping[UUID, UUID]
    post_scheduler: PostScheduler
    post_schedulers_params: Dict[str, Dict[str, Any]]
    score_scheduler: ScoreScheduler
    score_schedulers_params: Dict[str, Dict[str, Any]]
    iterations: int
    batch_size: Optional[int]
    lock: Lock


@dataclass
class FeedState:
    feed: List[PostData]
    length: int
    stream: Observable[PostData]


@dataclass
class State:
    face: FaceState
    module: ModuleState
    offline: OfflineState
    online: OnlineState
    training_task: Optional[asyncio.Task]
    training_status: Observable[TrainingStatus]
    feed: FeedState



import asyncio
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, AsyncIterable, Awaitable, Callable, Dict

from dateutil.parser import isoparse
from kilroy_server_py_utils import Configurable, Parameter, classproperty

from cilroy.models import SerializableState
from cilroy.scoring import ScoreScheduler
from cilroy.utils import next_time, seconds_until, utcmidnight


class State(SerializableState):
    base: datetime = utcmidnight()
    interval: timedelta = timedelta(days=1)


class IntervalScoreScheduler(ScoreScheduler, Configurable[State]):
    class BaseParameter(Parameter[State, str]):
        async def _get(self, state: State) -> str:
            return state.base.isoformat().replace("+00:00", "Z")

        async def _set(
            self, state: State, value: str
        ) -> Callable[[], Awaitable]:
            original_value = state.base

            async def undo():
                state.base = original_value

            state.base = isoparse(value)
            return undo

        @classproperty
        def schema(cls) -> Dict[str, Any]:
            return {
                "type": "string",
                "format": "date-time",
                "title": cls.pretty_name,
                "default": utcmidnight().isoformat().replace("+00:00", "Z"),
            }

    class IntervalParameter(Parameter[State, float]):
        async def _get(self, state: State) -> float:
            return state.interval.total_seconds()

        async def _set(
            self, state: State, value: float
        ) -> Callable[[], Awaitable]:
            original_value = state.interval

            async def undo():
                state.interval = original_value

            state.interval = timedelta(seconds=value)
            return undo

        @classproperty
        def schema(cls) -> Dict[str, Any]:
            return {
                "type": "number",
                "minimum": 0,
                "title": cls.pretty_name,
                "default": timedelta(days=1).total_seconds(),
            }

    @classmethod
    async def _save_state(cls, state: State, directory: Path) -> None:
        with open(directory / "state.json", "w") as f:
            f.write(state.json())

    async def _load_saved_state(self, directory: Path) -> State:
        with open(directory / "state.json", "r") as f:
            return State.parse_raw(f.read())

    async def wait(self) -> AsyncIterable[None]:
        while True:
            async with self.state.read_lock() as state:
                score_time = next_time(state.base, state.interval)

            await asyncio.sleep(seconds_until(score_time))
            yield



from dataclasses import dataclass
from typing import List

import betterproto


class Status(betterproto.Enum):
    STATUS_UNSPECIFIED = 0
    STATUS_LOADING = 1
    STATUS_READY = 2


class TrainingStatus(betterproto.Enum):
    TRAINING_STATUS_UNSPECIFIED = 0
    TRAINING_STATUS_IDLE = 1
    TRAINING_STATUS_OFFLINE = 2
    TRAINING_STATUS_ONLINE = 3


@dataclass(eq=False, repr=False)
class GetFaceMetadataRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetFaceMetadataResponse(betterproto.Message):
    key: str = betterproto.string_field(1)
    description: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class GetModuleMetadataRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetModuleMetadataResponse(betterproto.Message):
    key: str = betterproto.string_field(1)
    description: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class GetFacePostSchemaRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetFacePostSchemaResponse(betterproto.Message):
    schema: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetModulePostSchemaRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetModulePostSchemaResponse(betterproto.Message):
    schema: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetControllerStatusRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetControllerStatusResponse(betterproto.Message):
    status: "Status" = betterproto.enum_field(1)


@dataclass(eq=False, repr=False)
class WatchControllerStatusRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class WatchControllerStatusResponse(betterproto.Message):
    status: "Status" = betterproto.enum_field(1)


@dataclass(eq=False, repr=False)
class GetFaceStatusRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetFaceStatusResponse(betterproto.Message):
    status: "Status" = betterproto.enum_field(1)


@dataclass(eq=False, repr=False)
class WatchFaceStatusRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class WatchFaceStatusResponse(betterproto.Message):
    status: "Status" = betterproto.enum_field(1)


@dataclass(eq=False, repr=False)
class GetModuleStatusRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetModuleStatusResponse(betterproto.Message):
    status: "Status" = betterproto.enum_field(1)


@dataclass(eq=False, repr=False)
class WatchModuleStatusRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class WatchModuleStatusResponse(betterproto.Message):
    status: "Status" = betterproto.enum_field(1)


@dataclass(eq=False, repr=False)
class GetTrainingStatusRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetTrainingStatusResponse(betterproto.Message):
    status: "TrainingStatus" = betterproto.enum_field(1)


@dataclass(eq=False, repr=False)
class WatchTrainingStatusRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class WatchTrainingStatusResponse(betterproto.Message):
    status: "TrainingStatus" = betterproto.enum_field(1)


@dataclass(eq=False, repr=False)
class GetControllerConfigSchemaRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetControllerConfigSchemaResponse(betterproto.Message):
    schema: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetControllerConfigRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetControllerConfigResponse(betterproto.Message):
    config: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class WatchControllerConfigRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class WatchControllerConfigResponse(betterproto.Message):
    config: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class SetControllerConfigRequest(betterproto.Message):
    config: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class SetControllerConfigResponse(betterproto.Message):
    config: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetFaceConfigSchemaRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetFaceConfigSchemaResponse(betterproto.Message):
    schema: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetFaceConfigRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetFaceConfigResponse(betterproto.Message):
    config: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class WatchFaceConfigRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class WatchFaceConfigResponse(betterproto.Message):
    config: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class SetFaceConfigRequest(betterproto.Message):
    config: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class SetFaceConfigResponse(betterproto.Message):
    config: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetModuleConfigSchemaRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetModuleConfigSchemaResponse(betterproto.Message):
    schema: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetModuleConfigRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetModuleConfigResponse(betterproto.Message):
    config: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class WatchModuleConfigRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class WatchModuleConfigResponse(betterproto.Message):
    config: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class SetModuleConfigRequest(betterproto.Message):
    config: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class SetModuleConfigResponse(betterproto.Message):
    config: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class TrainOfflineRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class TrainOfflineResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class TrainOnlineRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class TrainOnlineResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class StopTrainingRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class StopTrainingResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class MetricConfig(betterproto.Message):
    id: str = betterproto.string_field(1)
    label: str = betterproto.string_field(2)
    group: str = betterproto.string_field(3)
    config: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class GetModuleMetricsConfigRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetModuleMetricsConfigResponse(betterproto.Message):
    configs: List["MetricConfig"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class MetricData(betterproto.Message):
    metric_id: str = betterproto.string_field(1)
    dataset_id: int = betterproto.uint64_field(2)
    data: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class GetModuleMetricsRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetModuleMetricsResponse(betterproto.Message):
    metrics: List["MetricData"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class WatchModuleMetricsRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class WatchModuleMetricsResponse(betterproto.Message):
    metric: "MetricData" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class WatchAllRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class WatchAllResponse(betterproto.Message):
    method: str = betterproto.string_field(1)
    message: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ResetControllerRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ResetControllerResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ResetFaceRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ResetFaceResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ResetModuleRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ResetModuleResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class Post(betterproto.Message):
    id: str = betterproto.string_field(1)
    url: str = betterproto.string_field(2)
    content: str = betterproto.string_field(3)
    created_at: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class GetFeedRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetFeedResponse(betterproto.Message):
    posts: List["Post"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class WatchFeedRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class WatchFeedResponse(betterproto.Message):
    post: "Post" = betterproto.message_field(1)
